@@ -62,7 +62,7 @@ int main()
    v[1] = 1;
    v[0] = v[1];
 
-   if (v[0] != v[1]) Error("BitMatTest not OK!!");
+   if (v[0] != v[1]) TerminalError("BitMatTest not OK!!");
 
    for (i=0; i < 8; i++) {
       mat_zz_p a, x;
@@ -77,22 +77,18 @@ int main()
       random(a, n, m);
 
       t = GetTime();
-      kernel(x, a);
+      image(x, a);
       t = GetTime() - t;  cerr << t << "\n";
 
       cvt(A, a);
 
       t = GetTime();
-      kernel(X, A);
+      image(X, A);
       t = GetTime() - t;  cerr << t << "\n";
 
-      cerr << x.NumRows() << "\n";
-
       cvt(X1, x);
 
-      if (X1 != X) Error("BitMatTest NOT OK!!");
-
-      if (!IsZero(X*A)) Error("BitMatTest NOT OK!!");
+      if (X1 != X) TerminalError("BitMatTest NOT OK!!");
 
       cerr << "\n";
    }
