@@ -3,7 +3,7 @@
 #include <NTL/vec_long.h>
 
 #include <NTL/new.h>
-#include <stdio.h>
+#include <cstdio>
 
 #if (defined(NTL_WIZARD_HACK) && defined(NTL_GF2X_LIB))
 #undef NTL_GF2X_LIB
@@ -13,16 +13,42 @@
 #include <gf2x.h>
 #endif
 
+
+#ifdef NTL_PCLMUL
+
+#if (NTL_BITS_PER_LONG != 64)
+#error "NTL_PCLMUL only works on 64-bit machines"
+#endif
+
+#include <wmmintrin.h>
+
+#define NTL_INLINE inline
+
+static inline void
+pclmul_mul1 (unsigned long *c, unsigned long a, unsigned long b)
+{
+   __m128i aa = _mm_setr_epi64( _mm_cvtsi64_m64(a), _mm_cvtsi64_m64(0));
+   __m128i bb = _mm_setr_epi64( _mm_cvtsi64_m64(b), _mm_cvtsi64_m64(0));
+   _mm_storeu_si128((__m128i*)c, _mm_clmulepi64_si128(aa, bb, 0));
+}
+#else
+
+
+#define NTL_INLINE
+
+#endif
+
 NTL_START_IMPL
 
+NTL_CHEAP_THREAD_LOCAL
 long GF2X::HexOutput = 0;
 
 
 void GF2X::SetMaxLength(long n)
 {
-   if (n < 0) Error("GF2X::SetMaxLength: negative length");
+   if (n < 0) LogicError("GF2X::SetMaxLength: negative length");
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("GF2X::SetMaxLength: excessive length");
+      ResourceError("GF2X::SetMaxLength: excessive length");
    long w = (n + NTL_BITS_PER_LONG - 1)/NTL_BITS_PER_LONG;
    xrep.SetMaxLength(w);
 }
@@ -36,7 +62,7 @@ GF2X::GF2X(INIT_SIZE_TYPE, long n)
 
 const GF2X& GF2X::zero()
 {
-   NTL_THREAD_LOCAL static GF2X z;
+   static const GF2X z; // GLOBAL (assumes C++11 thread-safe init)
    return z;
 }
 
@@ -59,8 +85,8 @@ void GF2X::normalize()
 void GF2X::SetLength(long n)
 {
    if (n < 0) {
-      Error("SetLength: negative index");
-      return;
+      LogicError("SetLength: negative index");
+      return; // NOTE: this helps the compiler optimize
    }
 
    long w = (n + NTL_BITS_PER_LONG - 1)/NTL_BITS_PER_LONG;
@@ -92,9 +118,9 @@ void GF2X::SetLength(long n)
 
 ref_GF2 GF2X::operator[](long i)
 {
-   if (i < 0) Error("GF2X: subscript out of range");
+   if (i < 0) LogicError("GF2X: subscript out of range");
    long wi = i/NTL_BITS_PER_LONG;
-   if (wi >= xrep.length())  Error("GF2X: subscript out of range");
+   if (wi >= xrep.length())  LogicError("GF2X: subscript out of range");
    long bi = i - wi*NTL_BITS_PER_LONG;
    return ref_GF2(INIT_LOOP_HOLE, &xrep[wi], bi);
 }
@@ -102,9 +128,9 @@ ref_GF2 GF2X::operator[](long i)
 
 const GF2 GF2X::operator[](long i) const
 {
-   if (i < 0) Error("GF2X: subscript out of range");
+   if (i < 0) LogicError("GF2X: subscript out of range");
    long wi = i/NTL_BITS_PER_LONG;
-   if (wi >= xrep.length())  Error("GF2X: subscript out of range");
+   if (wi >= xrep.length())  LogicError("GF2X: subscript out of range");
    long bi = i - wi*NTL_BITS_PER_LONG;
    return to_GF2((xrep[wi] & (1UL << bi)) != 0);
 }
@@ -172,8 +198,8 @@ void SetX(GF2X& x)
 void SetCoeff(GF2X& x, long i)
 {
    if (i < 0) {
-      Error("SetCoeff: negative index");
-      return;
+      LogicError("SetCoeff: negative index");
+      return; // NOTE: helpse the compiler optimize
    }
 
    long n, j;
@@ -197,8 +223,8 @@ void SetCoeff(GF2X& x, long i)
 void SetCoeff(GF2X& x, long i, long val)
 {
    if (i < 0) {
-      Error("SetCoeff: negative index");
-      return;
+      LogicError("SetCoeff: negative index");
+      return; // NOTE: helps the compiler optimize
    }
 
    val = val & 1;
@@ -230,12 +256,6 @@ void SetCoeff(GF2X& x, long i, GF2 a)
 }
 
 
-void swap(GF2X& a, GF2X& b)
-{
-   swap(a.xrep, b.xrep);
-}
-
-
 
 long deg(const GF2X& aa)
 {
@@ -247,7 +267,7 @@ long deg(const GF2X& aa)
    _ntl_ulong a = aa.xrep[n-1];
    long i = 0;
 
-   if (a == 0) Error("GF2X: unnormalized polynomial detected in deg");
+   if (a == 0) LogicError("GF2X: unnormalized polynomial detected in deg");
 
    while (a>=256)
       i += 8, a >>= 8;
@@ -325,7 +345,7 @@ istream & operator>>(istream& s, GF2X& a)
    NTL_ZZRegister(ival);
 
    long c;   
-   if (!s) Error("bad GF2X input"); 
+   if (!s) NTL_INPUT_ERROR(s, "bad GF2X input"); 
    
    c = s.peek();  
    while (IsWhiteSpace(c)) {  
@@ -341,12 +361,12 @@ istream & operator>>(istream& s, GF2X& a)
          return HexInput(s, a);
       }
       else {
-         Error("bad GF2X input");
+         NTL_INPUT_ERROR(s, "bad GF2X input");
       }
    }
 
    if (c != '[') {  
-      Error("bad GF2X input");  
+      NTL_INPUT_ERROR(s, "bad GF2X input");  
    }  
 
    GF2X ibuf;  
@@ -363,7 +383,7 @@ istream & operator>>(istream& s, GF2X& a)
    }  
 
    while (c != ']' && c != EOF) {   
-      if (!(s >> ival)) Error("bad GF2X input");
+      if (!(s >> ival)) NTL_INPUT_ERROR(s, "bad GF2X input");
       SetCoeff(ibuf, n, to_GF2(ival));
       n++;
 
@@ -375,7 +395,7 @@ istream & operator>>(istream& s, GF2X& a)
       }  
    }   
 
-   if (c == EOF) Error("bad GF2X input");  
+   if (c == EOF) NTL_INPUT_ERROR(s, "bad GF2X input");  
    s.get(); 
    
    a = ibuf; 
@@ -446,10 +466,10 @@ ostream& operator<<(ostream& s, const GF2X& a)
 
 void random(GF2X& x, long n)
 {
-   if (n < 0) Error("GF2X random: negative length");
+   if (n < 0) LogicError("GF2X random: negative length");
 
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("GF2X random: excessive length");
+      ResourceError("GF2X random: excessive length");
 
    long wl = (n+NTL_BITS_PER_LONG-1)/NTL_BITS_PER_LONG;
 
@@ -563,7 +583,11 @@ static
 void mul1(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
 {
 
+#ifdef NTL_PCLMUL
+pclmul_mul1(c, a, b);
+#else
 NTL_EFF_BB_MUL_CODE0
+#endif
 
 
 }
@@ -579,8 +603,11 @@ static inline
 void mul1_inline(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
 {
 
-
+#ifdef NTL_PCLMUL
+pclmul_mul1(c, a, b);
+#else
 NTL_EFF_BB_MUL_CODE0
+#endif
 
 
 }
@@ -593,9 +620,26 @@ static
 void Mul1(_ntl_ulong *cp, const _ntl_ulong *bp, long sb, _ntl_ulong a)
 {
  
+#ifdef NTL_PCLMUL
+
+   long i;
+   unsigned long carry, prod[2];
+
+   carry = 0;
+   for (i = 0; i < sb; i++) {
+      pclmul_mul1(prod, bp[i], a);
+      cp[i] = carry ^ prod[0];
+      carry = prod[1];
+   }
+
+   cp[sb] = carry;
+
+#else
 
 NTL_EFF_BB_MUL_CODE1
 
+#endif
+
 
 }
 
@@ -603,9 +647,27 @@ static
 void AddMul1(_ntl_ulong *cp, const _ntl_ulong* bp, long sb, _ntl_ulong a)
 {
 
+#ifdef NTL_PCLMUL
+
+   long i;
+   unsigned long carry, prod[2];
+
+   carry = 0;
+   for (i = 0; i < sb; i++) {
+      pclmul_mul1(prod, bp[i], a);
+      cp[i] ^= carry ^ prod[0];
+      carry = prod[1];
+   }
+
+   cp[sb] ^= carry;
+
+#else
 
 NTL_EFF_BB_MUL_CODE2
 
+#endif
+
+
 
 }
 
@@ -614,9 +676,26 @@ static
 void Mul1_short(_ntl_ulong *cp, const _ntl_ulong *bp, long sb, _ntl_ulong a)
 {
  
+#ifdef NTL_PCLMUL
+
+   long i;
+   unsigned long carry, prod[2];
+
+   carry = 0;
+   for (i = 0; i < sb; i++) {
+      pclmul_mul1(prod, bp[i], a);
+      cp[i] = carry ^ prod[0];
+      carry = prod[1];
+   }
+
+   cp[sb] = carry;
+
+#else
 
 NTL_EFF_SHORT_BB_MUL_CODE1
 
+#endif
+
 
 }
 
@@ -628,8 +707,11 @@ static
 void mul_half(_ntl_ulong *c, _ntl_ulong a, _ntl_ulong b)
 {
 
-
+#ifdef NTL_PCLMUL
+pclmul_mul1(c, a, b);
+#else
 NTL_EFF_HALF_BB_MUL_CODE0
+#endif
 
 
 }
@@ -637,9 +719,12 @@ NTL_EFF_HALF_BB_MUL_CODE0
 
 // mul2...mul8 hard-code 2x2...8x8 word multiplies.
 // I adapted these routines from LiDIA (except mul3, see below).
-// Inlining these seems to hurt, not help.
+// NOTE: Generally, inlining these functions seems to hurt performance,
+// at least when using software mul1; for hardware mul1, I've
+// switched to making mul2, 3, 4 inline, although this should
+// really be profiled.
 
-static
+static NTL_INLINE
 void mul2(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
 {
    _ntl_ulong hs0, hs1;
@@ -667,7 +752,7 @@ void mul2(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
  * an its application to ECC" (ACISP 2003).
  */
 
-static
+static NTL_INLINE
 void mul3 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
 {
    _ntl_ulong d0[2], d1[2], d2[2], d01[2], d02[2], d12[2];
@@ -689,7 +774,7 @@ void mul3 (_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
 
 }
 
-static
+static NTL_INLINE
 void mul4(_ntl_ulong *c, const _ntl_ulong *a, const _ntl_ulong *b)
 {
    _ntl_ulong hs0[2], hs1[2];
@@ -953,89 +1038,11 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
       return;
    }
 
-   if (sa == 1 && sb == 1) {
-      // special case...
-      _ntl_ulong v[2];
-      if (!(a0 >> NTL_BITS_PER_LONG/2))
-         mul_half(v, b0, a0);
-      else if (!(b0 >> NTL_BITS_PER_LONG/2))
-         mul_half(v, a0, b0);
-      else
-         mul1(v, a0, b0);
-
-      if (v[1]) {
-         c.xrep.SetLength(2);
-         _ntl_ulong *cp = &c.xrep[0];
-         cp[0] = v[0];
-         cp[1] = v[1];
-      }
-      else {
-         c.xrep.SetLength(1);
-         _ntl_ulong *cp = &c.xrep[0];
-         cp[0] = v[0];
-      }
-      return;
-   }
-
-   if (sa == 2 && sb == 2) {
-      // special case...
-      _ntl_ulong v[4];
-      mul2(v, &a.xrep[0], &b.xrep[0]);
-      if (v[3]) {
-         c.xrep.SetLength(4);
-         _ntl_ulong *cp = &c.xrep[0];
-         cp[0] = v[0];
-         cp[1] = v[1];
-         cp[2] = v[2];
-         cp[3] = v[3];
-      }
-      else {
-         c.xrep.SetLength(3);
-         _ntl_ulong *cp = &c.xrep[0];
-         cp[0] = v[0];
-         cp[1] = v[1];
-         cp[2] = v[2];
-      }
-      return;
-   }
-
-   // another special case:  one of the two inputs
-   // has length 1 (or less).
-
-   if (sa == 1) {
-      c.xrep.SetLength(sb + 1);
-      _ntl_ulong *cp = c.xrep.elts();
-      const _ntl_ulong *bp = b.xrep.elts();
-
-      if (a0 >> (NTL_BITS_PER_LONG-NTL_BB_MUL1_BITS+1))
-         Mul1(cp, bp, sb, a0);
-      else
-         Mul1_short(cp, bp, sb, a0);
-
-
-      c.normalize();
-      return;
-   }
-
-   if (sb == 1) {
-      c.xrep.SetLength(sa + 1);
-      _ntl_ulong *cp = c.xrep.elts();
-      const _ntl_ulong *ap = a.xrep.elts();
-
-
-      if (b0 >> (NTL_BITS_PER_LONG-NTL_BB_MUL1_BITS+1))
-         Mul1(cp, ap, sa, b0);
-      else
-         Mul1_short(cp, ap, sa, b0);
-
-      c.normalize();
-      return;
-   }
-
    // finally: the general case
 
 
-   NTL_THREAD_LOCAL static WordVector mem;
+   NTL_TLS_LOCAL(WordVector, mem);
+   WordVectorWatcher watch_mem(mem);
 
    const _ntl_ulong *ap = a.xrep.elts(), *bp = b.xrep.elts();
    _ntl_ulong *cp;
@@ -1061,7 +1068,6 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
    }
 
    c.normalize();
-   mem.release();
 }
 #else
 void OldMul(GF2X& c, const GF2X& a, const GF2X& b)
@@ -1401,9 +1407,13 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
    // finally: the general case
 
    
-   NTL_THREAD_LOCAL static WordVector mem;
-   NTL_THREAD_LOCAL static WordVector stk;
-   NTL_THREAD_LOCAL static WordVector vec;
+   NTL_TLS_LOCAL(WordVector, mem);
+   NTL_TLS_LOCAL(WordVector, stk);
+   NTL_TLS_LOCAL(WordVector, vec);
+
+   WordVectorWatcher watch_mem(mem);
+   WordVectorWatcher watch_stk(stk);
+   WordVectorWatcher watch_vec(vec);
 
    const _ntl_ulong *ap, *bp;
    _ntl_ulong *cp;
@@ -1482,10 +1492,6 @@ void mul(GF2X& c, const GF2X& a, const GF2X& b)
       c.xrep = mem;
 
    c.normalize();
-
-   mem.release();
-   stk.release();
-   vec.release();
 }
 
 
@@ -1509,7 +1515,7 @@ void mul(GF2X& c, const GF2X& a, GF2 b)
 
 void trunc(GF2X& x, const GF2X& a, long m)
 {
-   if (m < 0) Error("trunc: bad args");
+   if (m < 0) LogicError("trunc: bad args");
 
    long n = a.xrep.length();
    if (n == 0 || m == 0) {
@@ -1620,15 +1626,22 @@ static const _ntl_ulong sqrtab[256] = {
 
 
 
-static inline 
+
+static inline
 void sqr1(_ntl_ulong *c, _ntl_ulong a)
 {
+#ifdef NTL_PCLMUL
+   // this appears to be marginally faster than the
+   // table-driven code
+   pclmul_mul1(c, a, a);
+#else
    _ntl_ulong hi, lo;
 
    NTL_BB_SQR_CODE
 
    c[0] = lo;
    c[1] = hi;
+#endif
 }
 
 
@@ -1677,7 +1690,7 @@ void LeftShift(GF2X& c, const GF2X& a, long n)
    }
 
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("overflow in LeftShift");
+      ResourceError("overflow in LeftShift");
 
    if (n == 0) {
       c = a;
@@ -1720,7 +1733,7 @@ void LeftShift(GF2X& c, const GF2X& a, long n)
 void ShiftAdd(GF2X& c, const GF2X& a, long n)
 // c = c + a*X^n
 {
-   if (n < 0) Error("ShiftAdd: negative argument");
+   if (n < 0) LogicError("ShiftAdd: negative argument");
 
    if (n == 0) {
       add(c, c, a);
@@ -1728,7 +1741,7 @@ void ShiftAdd(GF2X& c, const GF2X& a, long n)
    }
 
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("overflow in ShiftAdd");
+      ResourceError("overflow in ShiftAdd");
 
    long sa = a.xrep.length();
    if (sa <= 0) {
@@ -1779,7 +1792,7 @@ void RightShift(GF2X& c, const GF2X& a, long n)
    }
 
    if (n < 0) {
-      if (n < -NTL_MAX_LONG) Error("overflow in RightShift");
+      if (n < -NTL_MAX_LONG) ResourceError("overflow in RightShift");
       LeftShift(c, a, -n);
       return;
    }
@@ -1865,7 +1878,7 @@ void CopyReverse(GF2X& c, const GF2X& a, long hi)
    if (hi < 0) { clear(c); return; }
 
    if (NTL_OVERFLOW(hi, 1, 0))
-      Error("overflow in CopyReverse");
+      ResourceError("overflow in CopyReverse");
 
    long n = hi+1;
    long sa = a.xrep.length();
@@ -1916,7 +1929,7 @@ void CopyReverse(GF2X& c, const GF2X& a, long hi)
 void div(GF2X& q, const GF2X& a, GF2 b)
 {
    if (b == 0)
-      Error("div: division by zero");
+      ArithmeticError("div: division by zero");
 
    q = a;
 }
@@ -1924,7 +1937,7 @@ void div(GF2X& q, const GF2X& a, GF2 b)
 void div(GF2X& q, const GF2X& a, long b)
 {
    if ((b & 1) == 0)
-      Error("div: division by zero");
+      ArithmeticError("div: division by zero");
 
    q = a;
 }
